unit TransBtn;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, Buttons, ExtCtrls,
  Math, Dialogs;

type
  TTranslateMode = (tmScl, tmShr);
  TTransBtn = class(TButton)
  private
    FBackBeforeHoverColor: TColor;
  private
    FRotCntPt :TPoint;
    FRgnImg :TImage;
    FDeg, FSclX, FSclY :Real;
    FMaskColor :TColor;
    FRgnBox :TRect;
    FRgnRop :Cardinal;

    FCanvas: TCanvas;
    IsFocused: Boolean;
    FBackColor: TColor;
    FForeColor: TColor;
    FHoverColor: TColor;

    procedure SetBackColor(const Value: TColor);
    procedure SetForeColor(const Value: TColor);
    procedure SetHoverColor(const Value: TColor);
    property BackBeforeHoverColor : TColor read FBackBeforeHoverColor write FBackBeforeHoverColor;

    function  GetRgnBmp:TBitmap;
    procedure SetRgnBmp(Value: TBitmap);

    function  GetRgnSclX:Real;
    procedure SetRgnSclX(Value: Real);
    function  GetRgnSclY:Real;
    procedure SetRgnSclY(Value: Real);
    function  GetRgnDeg: Real;
    procedure SetRgnDeg(Value: Real);
    function  GetRgnMaskColor: TColor;
    procedure SetRgnMaskColor(Value: TColor);
    function  GetRgnCntPos: TPoint;
    procedure SetRgnCntPos(Value: TPoint);
    function  GetRgnRop: Cardinal;
    procedure SetRgnRop(Value: Cardinal);
  private
    function GetRotationXForm(Deg :Single; CntPt, OffsetPt :TPoint):TXForm;
    function GetXForm(VarX, VarY :Single; OffsetPt :TPoint; Mode :TTranslateMode):TXForm;
    function CreateTxtRgn(Font :TFont; Txt :String):HRGN;
    function TranslateRgn(Rgn :HRGN; TXF :TXForm):HRGN;
    function CreateRgnFromBmp(bmp: TBitmap; TransparentColor: TColor =clBlack; RedTol: Byte =1; GreenTol: Byte =1; BlueTol: Byte =1):HRGN;
    procedure SetCanvasZoom(dc: HDC; ZoomX, ZoomY: Single);
    function TranslateWndFromPicture(TGTWnd :HWND; Picture :TPicture; RotCntPt :TPoint; clMask :TColor=clBlack; deg :Real=0; SclX :Real=1.0; SclY :Real=1.0; Rop :Cardinal=SRCCOPY):TRect;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Message : TMessage); override;

    procedure SetButtonStyle(Value: Boolean); override;
    procedure Resize; override;
    procedure DrawButton(Rect: TRect; State: UINT);
    procedure DrawNoPictureBtn(Rect: TRect; State: UINT);

    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CNMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;

    procedure OnMove(var Msg: TWMMove); message WM_MOVE;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadRgnPicture(Const Filename :string);
  published
    property BackColor: TColor read FBackColor write SetBackColor default clBtnFace;
    property ForeColor: TColor read FForeColor write SetForeColor default clBtnText;
    property HoverColor: TColor read FHoverColor write SetHoverColor default clBtnFace;
    property RgnBmp: TBitmap read GetRgnBmp Write SetRgnBmp;
    Property RgnSclX :Real read GetRgnSclX write SetRgnSclX ;
    Property RgnSclY :Real read GetRgnSclY write SetRgnSclY ;
    Property RgnDeg :Real read GetRgnDeg write SetRgnDeg ;
    Property RgnMaskColor :TColor read GetRgnMaskColor write SetRgnMaskColor default clBlack;
    Property RgnCntPos :TPoint read GetRgnCntPos write SetRgnCntPos;
    Property RgnRop :Cardinal read GetRgnRop write SetRgnRop default SRCCOPY;
  end;
              
procedure Register;

implementation
//---------------------------------------------------------------------
constructor TTransBtn.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRgnImg :=TImage.Create(nil);
  FDeg :=0;
  FSclX :=1.0;
  FSclY :=1.0;;
  FMaskColor :=clBlack;
  FRotCntPt :=Point(Self.Width div 2, Self.Height div 2);
  FRgnRop :=SRCCOPY;

  FCanvas := TCanvas.Create;
  BackColor := clBtnFace;
  ForeColor := clBtnText;
  HoverColor := clBtnFace;

end; (*Create*)

destructor TTransBtn.Destroy;
begin
  FCanvas.Free;
  FRgnImg.Free;
  inherited Destroy;
end; (*Destroy*)

procedure TTransBtn.WndProc(var Message : TMessage);
begin
  if (Message.Msg = CM_MOUSELEAVE) then
  begin
    BackColor := BackBeforeHoverColor;
    invalidate;
  end;
  if (Message.Msg = CM_MOUSEENTER) then
  begin
    BackBeforeHoverColor := BackColor;
    BackColor := HoverColor;
    invalidate;
  end;
  inherited;
end; (*WndProc*)

procedure TTransBtn.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or BS_OWNERDRAW;
end; (*CreateParams*)

procedure TTransBtn.SetButtonStyle(Value: Boolean);
begin
  if Value <> IsFocused then
  begin
    IsFocused := Value;
    Invalidate;
  end;
end; (*SetButtonStyle*)

procedure TTransBtn.CNMeasureItem(var Message: TWMMeasureItem);
begin
  with Message.MeasureItemStruct^ do
  begin
    itemWidth  := Width; 
    itemHeight := Height; 
  end; 
end; (*CNMeasureItem*)

procedure TTransBtn.CNDrawItem(var Message: TWMDrawItem);
var 
  SaveIndex: Integer;
begin
  with Message.DrawItemStruct^ do
  begin
    SaveIndex := SaveDC(hDC);
    FCanvas.Lock; 
    try 
      FCanvas.Handle := hDC;
      FCanvas.Font := Font;
      FCanvas.Brush := self. Brush;
      DrawButton(rcItem, itemState);
    finally
      FCanvas.Handle := 0;
      FCanvas.Unlock;
      RestoreDC(hDC, SaveIndex);
    end;
  end; 
  Message.Result := 1; 
end; (*CNDrawItem*)

procedure TTransBtn.OnMove(var Msg: TWMMove);
begin
  FRotCntPt :=Point(Self.Left + Self.Width div 2, Self.Top + Self.Height div 2);
  inherited;
end;

procedure TTransBtn.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end; (*CMEnabledChanged*)

procedure TTransBtn.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end; (*CMFontChanged*)


procedure TTransBtn.SetBackColor(const Value: TColor);
begin
  if FBackColor <> Value then 
  begin
    FBackColor:= Value;
    Invalidate;
  end;
end; (*SetButtonColor*)

procedure TTransBtn.SetForeColor(const Value: TColor);
begin
  if FForeColor <> Value then 
  begin
    FForeColor:= Value;
    Invalidate;
  end;
end; (*SetForeColor*)

procedure TTransBtn.SetHoverColor(const Value: TColor);
begin
  if FHoverColor <> Value then 
  begin
    FHoverColor:= Value;
    Invalidate;
  end;
end; (*SetHoverColor*)

procedure TTransBtn.DrawNoPictureBtn(Rect: TRect; State: UINT);
var
  Flags, OldMode: Longint;
  IsDown, IsDefault, IsDisabled: Boolean;
  OldColor: TColor;
  OrgRect: TRect;
begin
  OrgRect := Rect;
  Flags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT;
  IsDown := State and ODS_SELECTED <> 0;
  IsDisabled := State and ODS_DISABLED <> 0;
  IsDefault := State and ODS_FOCUS <> 0;

  if IsDown then Flags := Flags or DFCS_PUSHED;
  if IsDisabled then Flags := Flags or DFCS_INACTIVE;

  if (IsFocused or IsDefault) then 
  begin 
    FCanvas.Pen.Color := clWindowFrame; 
    FCanvas.Pen.Width := 1; 
    FCanvas.Brush.Style := bsClear; 
    FCanvas.Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom); 
    InflateRect(Rect, - 1, - 1); 
  end; 

  if IsDown then 
  begin
    FCanvas.Pen.Color := clBtnShadow;
    FCanvas.Pen.Width := 1; 
    FCanvas.Brush.Color := clBtnFace; 
    FCanvas.Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom); 
    InflateRect(Rect, - 1, - 1); 
  end 
  else
  begin 
    DrawFrameControl(FCanvas.Handle, Rect, DFC_BUTTON, Flags);
  end; 

  if IsDown then OffsetRect(Rect, 1, 1); 

  OldColor := FCanvas.Brush.Color;
  FCanvas.Brush.Color := BackColor;
  FCanvas.FillRect(Rect); 
  FCanvas.Brush.Color := OldColor;
  OldMode := SetBkMode(FCanvas.Handle, TRANSPARENT); 
  FCanvas.Font.Color := ForeColor;
  if IsDisabled then
    DrawState(FCanvas.Handle, FCanvas.Brush.Handle, nil, Integer(Caption), 0, 
    ((Rect.Right - Rect.Left) - FCanvas.TextWidth(Caption)) div 2, 
    ((Rect.Bottom - Rect.Top) - FCanvas.TextHeight(Caption)) div 2, 
      0, 0, DST_TEXT or DSS_DISABLED) 
  else 
    DrawText(FCanvas.Handle, PChar(Caption), - 1, Rect, 
      DT_SINGLELINE or DT_CENTER or DT_VCENTER);
       
  SetBkMode(FCanvas.Handle, OldMode); 

  if (IsFocused and IsDefault) then
  begin
    Rect := OrgRect;
    InflateRect(Rect, - 4, - 4);
    FCanvas.Pen.Color := clWindowFrame;
    FCanvas.Brush.Color := clBtnFace;
    DrawFocusRect(FCanvas.Handle, Rect);
  end;
end; (*DrawNoPictureBtn*)

procedure TTransBtn.DrawButton(Rect: TRect; State: UINT);
var
  Flags, OldMode: Longint;
  IsFocus, IsDown, IsDefault, IsDisabled: Boolean;
begin
  Flags := DFCS_BUTTONPUSH or DFCS_ADJUSTRECT or DFCS_BUTTONRADIOIMAGE;
  IsDown := State and ODS_SELECTED <> 0;
  IsDisabled := State and ODS_DISABLED <> 0;
  IsDefault := State and ODS_DEFAULT <> 0;
  IsFocus := State and ODS_FOCUS <> 0;

  if FRgnImg.Picture.Bitmap.Empty then
  begin
    DrawNoPictureBtn(Rect, State);
  end else begin
    if IsDown then
    begin
      FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX-0.05, FSclY-0.05, FRgnRop);
    end else begin
      FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
    end;
  end;

end; (*DrawButton*)

procedure TTransBtn.LoadRgnPicture(Const Filename :string);
begin
  FRgnImg.Picture.Bitmap.LoadFromFile(Filename);
  if Self.Parent <> nil then
    FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
end;

function TTransBtn.GetRgnCntPos: TPoint;
begin
  Result :=FRotCntPt;
end;

function TTransBtn.GetRgnSclX:Real;
begin
  Result :=FSclX;
end;

function TTransBtn.GetRgnSclY:Real;
begin
  Result :=FSclY;
end;

function TTransBtn.GetRgnDeg: Real;
begin
  Result :=FDeg;
end;

function TTransBtn.GetRgnMaskColor: TColor;
begin
   Result :=FMaskColor;
end;

function TTransBtn.GetRgnRop: Cardinal;
begin
   Result :=FRgnRop;
end;

function TTransBtn.GetRgnBmp: TBitmap;
begin
   Result :=FRgnImg.Picture.Bitmap;
end;

procedure TTransBtn.SetRgnBmp(Value: TBitmap);
begin

  FRgnImg.Picture.Bitmap.Assign(Value);
  if Self.Parent <> nil then
    FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);

end;

procedure TTransBtn.SetRgnCntPos(Value: TPoint);
begin
  FRotCntPt :=Value;
  Self.Left :=Value.X - Self.Width div 2;
  Self.Top :=Value.Y - Self.Height div 2;
end;

procedure TTransBtn.SetRgnDeg(Value: Real);
begin
  FDeg :=Value;
  FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
end;

procedure TTransBtn.SetRgnSclX(Value: Real);
begin
  FSclX :=Value;
  FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
end;

procedure TTransBtn.SetRgnSclY(Value: Real);
begin
  FSclY :=Value;
  FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
end;

procedure TTransBtn.SetRgnMaskColor(Value: TColor);
begin
  FMaskColor :=Value;
end;

procedure TTransBtn.SetRgnRop(Value: Cardinal);
begin
  FRgnRop :=Value;
  FRgnBox :=TranslateWndFromPicture(Self.Handle, FRgnImg.Picture, FRotCntPt, FMaskColor, FDeg, FSclX, FSclY, FRgnRop);
end;

function TTransBtn.TranslateWndFromPicture(TGTWnd :HWND; Picture :TPicture; RotCntPt :TPoint; clMask :TColor=clBlack; deg :Real=0; SclX :Real=1.0; SclY :Real=1.0; Rop :Cardinal=SRCCOPY):TRect;
var
  BmpBox, WndBox, RgnBox :TRect;
  BmpW, BmpH, cx,cy :Integer;
  TXF :TXForm;
  BmpCntPt, WndTopLeftPt, CntPt, OfxPt :TPoint;
  Rgn :HRGN;
  Cvs :TCanvas;
  ps :PAINTSTRUCT;
begin
  Result :=Rect(0,0,0,0);
  if Picture.Bitmap.Empty then Exit;
  Cvs :=TCanvas.Create;
  cvs.Handle :=GetWindowDC(TGTWnd);

  BmpW := Picture.Bitmap.Width ;
  BmpH := Picture.Bitmap.Height ;
  BmpBox.TopLeft :=Point(0, 0);
  CntPt :=Point(0, 0);
  OfxPt :=Point(0, 0);
  BmpBox.BottomRight :=Point(Round(BmpW * SclX), Round(BmpH * SclY));
  BmpCntPt :=Point(Round(BmpW * SclX * 0.5), Round(BmpH * SclY *0.5));

  Rgn :=CreateRgnFromBmp(Picture.Bitmap, clMask);
  TXF :=GetRotationXForm(deg, CntPt, OfxPt);
  CombineTransform(TXF,GetXForm(SclX, SclY, OfxPt, tmScl), GetRotationXForm(deg, CntPt, OfxPt));
  Rgn :=TranslateRgn(Rgn, TXF);
  GetRgnBox(Rgn, RgnBox);
  OffsetRgn(Rgn, -RgnBox.Left, -RgnBox.Top);
  GetRgnBox(Rgn, RgnBox);
  Result :=RgnBox;

  cx := abs(RgnBox.Right - RgnBox.Left);
  cy := abs(RgnBox.Bottom - RgnBox.Top);
  WndTopLeftPt :=Point(RotCntPt.X - (cx div 2) , RotCntPt.Y - (cy div 2));
  SetWindowPos(TGTWnd, HWND_TOP, WndTopLeftPt.X, WndTopLeftPt.Y, cx, cy, SWP_SHOWWINDOW);
  SetWindowRgn(TGTWnd, Rgn, True);

  SetGraphicsMode(cvs.Handle, GM_ADVANCED);
  GetwindowRect(TGTWnd, WndBox);
  CntPt :=Point(abs(WndBox.Right-WndBox.Left) div 2, abs(WndBox.Bottom-WndBox.Top) div 2);
  OfxPt.X :=-Round((BmpCntPt.X-CntPt.X) * Cos(Deg*PI/180) - (BmpCntPt.Y-CntPt.y) * Sin(Deg*PI/180));
  OfxPt.Y :=-Round((BmpCntPt.X-CntPt.X) * Sin(Deg*PI/180) + (BmpCntPt.Y-CntPt.y) * Cos(Deg*PI/180));
  TXF :=GetRotationXForm(deg, CntPt, OfxPt);
  SetWorldTransform(cvs.Handle, TXF);

  BeginPaint(TGTWnd, ps); // bثeW,eWI
  StretchBlt(cvs.Handle, 0, 0, BmpBox.Right-BmpBox.Left, BmpBox.Bottom-BmpBox.Top,        
             Picture.Bitmap.Canvas.Handle,0, 0, Picture.Bitmap.Width, Picture.Bitmap.Height, Rop);
  EndPaint(TGTWnd, ps);

  ReleaseDC(TGTWnd, cvs.Handle);
  cvs.Free;
  DeleteObject(Rgn);
end;

procedure TTransBtn.SetCanvasZoom(dc: HDC; ZoomX, ZoomY: Single);
const
  ZoomBase =100.;
var
  WndOrgPt, ViewOrgPt :TPoint;
  WndExtSize, ViewExtSize, ViewExtSizeOld, WndExtSizeOld :Size;
  x, y :Integer;
begin
  SetMapMode(dc, MM_ANISOTROPIC);//MM_ISOTROPIC
  x :=Round(ZoomX * ZoomBase);
  y :=Round(ZoomY * ZoomBase);
  SetWindowExtEx(dc, Round(ZoomBase), Round(ZoomBase), @WndExtSizeOld);
  SetViewportExtEx(dc, x , y, @ViewExtSizeOld);
end;

function TTransBtn.GetRotationXForm(Deg :Single; CntPt, OffsetPt :TPoint):TXForm;
begin
   Result.eM11 :=Cos(Deg*PI/180);
   Result.eM12 :=Sin(Deg*PI/180);
   Result.eM21 :=-Sin(Deg*PI/180);
   Result.eM22 :=Cos(Deg*PI/180);
   Result.eDx :=OffsetPt.X + CntPt.X - CntPt.X * Cos(Deg*PI/180) + CntPt.y * Sin(Deg*PI/180);
   Result.eDy :=OffsetPt.Y + CntPt.y - CntPt.y * Cos(Deg*PI/180) - CntPt.x * Sin(Deg*PI/180);
end;

function TTransBtn.GetXForm(VarX, VarY :Single; OffsetPt :TPoint; Mode :TTranslateMode):TXForm;
var
  TXF :TXForm;
begin
  TXF.eM11 :=1;
  TXF.eM12 :=0;
  TXF.eM21 :=0;
  TXF.eM22 :=1;
  TXF.eDx :=OffsetPt.X;
  TXF.eDy :=OffsetPt.Y;
  Result :=TXF;

  case Mode of
  tmScl: begin
    TXF.eM11 :=VarX;
    TXF.eM22 :=VarY;
  end;
  tmShr:  begin
    TXF.eM12 :=VarX;
    TXF.eM21 :=VarY;
  end; else
    TXF.eM11 :=0;
    TXF.eM12 :=0;
    TXF.eM21 :=0;
    TXF.eM22 :=0;
    TXF.eDx :=0;
    TXF.eDy :=0;
  end;
  Result :=TXF;
end;

function TTransBtn.CreateTxtRgn(Font :TFont; Txt :String):HRGN;
var
  ErrCode :cardinal;
  cvs :TCanvas;
begin
  Result :=0;
  cvs :=TCanvas.Create;
  cvs.Font :=Font;
  cvs.Handle :=CreateDC('DISPLAY', nil, nil, nil);
  SetBkMode(cvs.Handle, TRANSPARENT);
  BeginPath(cvs.Handle);
  cvs.TextOut(0, 0, Txt);
  EndPath(cvs.Handle);

  Result :=PathToRegion(cvs.Handle);
  if Result =0 then
  begin
    ErrCode :=GetLastError;
    case ErrCode of
    ERROR_CAN_NOT_COMPLETE:  begin
      ShowMessage('ERROR_CAN_NOT_COMPLETE');
    end;
    ERROR_INVALID_PARAMETER: begin
      ShowMessage('ERROR_INVALID_PARAMETER');
    end;
    ERROR_NOT_ENOUGH_MEMORY: begin
      ShowMessage('ERROR_NOT_ENOUGH_MEMORY');
    end; else
      ShowMessage(SysErrorMessage(ErrCode)) ;
    end;
  end;
  DeleteDC(cvs.Handle);
  cvs.Free;
end;

function TTransBtn.TranslateRgn(Rgn :HRGN; TXF :TXForm):HRGN;
var
  OrgRgnData, RgnData: PRgnData;
  RgnDataSize: Integer;
  ErrCode :cardinal;
begin
  Result :=0;
  RgnDataSize := GetRegionData(Rgn, 0, nil);
  GetMem(OrgRgnData, RgnDataSize);
  GetMem(RgnData, RgnDataSize);
  GetRegionData(Rgn, RgnDataSize, OrgRgnData);
  CopyMemory(RgnData, OrgRgnData, RgnDataSize);
  DeleteObject(Rgn);
  Result :=ExtCreateRegion(@TXF, RgnDataSize, RgnData^);
  if Result =0 then
  begin
    ErrCode :=GetLastError;
    case ErrCode of
    ERROR_CAN_NOT_COMPLETE:  begin
      ShowMessage('ERROR_CAN_NOT_COMPLETE');
    end;
    ERROR_INVALID_PARAMETER: begin
      ShowMessage('ERROR_INVALID_PARAMETER');
    end;
    ERROR_NOT_ENOUGH_MEMORY: begin
      ShowMessage('ERROR_NOT_ENOUGH_MEMORY');
    end; else
      ShowMessage(SysErrorMessage(ErrCode)) ;
    end;
  end;
  FreeMem(OrgRgnData, RgnDataSize);
  FreeMem(RgnData, RgnDataSize);
end;

function TTransBtn.CreateRgnFromBmp(bmp: TBitmap; TransparentColor: TColor =clBlack;
                          RedTol: Byte =1; GreenTol: Byte =1; BlueTol: Byte =1):HRGN;
const
  AllocUnit =100;
type
  PRectArray =^TRectArray;
  TRectArray =array[0..(MaxInt div SizeOf(TRect)) - 1] of TRect;
var
  pr: PRectArray;               // used to access the rects array of RgnData by index
  h: HRGN;                      // Handles to regions
  RgnData: PRgnData;            // Pointer to structure RGNDATA used to create regions
  lr, lg, lb, hr, hg, hb: Byte; // values for lowest and hightest trans. colors
  x, y, x0: Integer;            // coordinates of current rect of visible pixels
  b: PByteArray;                // used to easy the task of testing the byte pixels (R,G,B)
  ScanLinePtr: Pointer;         // Pointer to current ScanLine being scanned
  ScanLineInc: Integer;         // Offset to next bitmap scanline (can be negative)
  maxRects: Cardinal;           // Number of rects to realloc memory by chunks of AllocUnit
begin
  Result :=0;
  { Keep on hand lowest and highest values for the "transparent" pixels }
  lr :=GetRValue(TransparentColor);
  lg :=GetGValue(TransparentColor);
  lb :=GetBValue(TransparentColor);
  hr :=Min($FF, lr + RedTol);
  hg :=Min($FF, lg + GreenTol);
  hb :=Min($FF, lb + BlueTol);
  { ensures that the pixel format is 32-bits per pixel }
  bmp.PixelFormat := pf32bit;
  { alloc initial region data }
  maxRects :=AllocUnit;
  GetMem(RgnData, SizeOf(RGNDATAHEADER) + (SizeOf(TRect) * maxRects));
  try
    with RgnData^.rdh do
    begin
      dwSize :=SizeOf(RGNDATAHEADER);
      iType :=RDH_RECTANGLES;
      nCount :=0;
      nRgnSize :=0;
      SetRect(rcBound, MAXLONG, MAXLONG, 0, 0);
    end;
    { scan each bitmap row - the orientation doesn't matter (Bottom-up or not) }
    ScanLinePtr :=bmp.ScanLine[0];
    ScanLineInc :=Integer(bmp.ScanLine[1]) - Integer(ScanLinePtr);
    for y :=0 to bmp.Height - 1 do
    begin
      x :=0;
      while x < bmp.Width do
      begin
        x0 :=x;
        while x < bmp.Width do
        begin
          b :=@PByteArray(ScanLinePtr)[x * SizeOf(TRGBQuad)];
          // BGR-RGB: Windows 32bpp BMPs are made of BGRa quads (not RGBa)
          if(b[2] >= lr) and (b[2] <= hr) and
            (b[1] >= lg) and (b[1] <= hg) and
            (b[0] >= lb) and (b[0] <= hb) then
            Break; // pixel is transparent
          Inc(x);
        end;
        { test to see if we have a non-transparent area in the image }
        if x > x0 then
        begin
          { increase RgnData by AllocUnit rects if we exceeds maxRects }
          if RgnData^.rdh.nCount >= maxRects then
          begin
            Inc(maxRects, AllocUnit);
            ReallocMem(RgnData, SizeOf(RGNDATAHEADER) + (SizeOf(TRect) * MaxRects));
          end;
          { Add the rect (x0, y)-(x, y+1) as a new visible area in the region }
          pr :=@RgnData^.Buffer; // Buffer is an array of rects
          with RgnData^.rdh do
          begin
            SetRect(pr[nCount], x0, y, x, y + 1);
            { adjust the bound rectangle of the region if we are "out-of-bounds" }
            if x0 < rcBound.Left then
              rcBound.Left :=x0;
            if y < rcBound.Top then
              rcBound.Top :=y;
            if x > rcBound.Right then
              rcBound.Right :=x;
            if y + 1 > rcBound.Bottom then
              rcBound.Bottom :=y + 1;
            Inc(nCount);
          end;
        end; // if x > x0
        { Need to create the region by muliple calls to ExtCreateRegion, 'cause }
        { it will fail on Windows 98 if the number of rectangles is too large   }
        if RgnData^.rdh.nCount =2000 then
        begin
          h := ExtCreateRegion(nil, SizeOf(RGNDATAHEADER) + (SizeOf(TRect) * maxRects), RgnData^);
          if Result > 0 then
          begin // Expand the current region
            CombineRgn(Result, Result, h, RGN_OR);
            DeleteObject(h);
          end else // First region, assign it to Result
            Result :=h;
            RgnData^.rdh.nCount :=0;
            SetRect(RgnData^.rdh.rcBound, MAXLONG, MAXLONG, 0, 0);
        end;
        Inc(x);
      end; // scan every sample byte of the image
      Inc(Integer(ScanLinePtr), ScanLineInc);
    end;
    { need to call ExCreateRegion one more time because we could have left    }
    { a RgnData with less than 2000 rects, so it wasn't yet created/combined  }
    h :=ExtCreateRegion(nil, SizeOf(RGNDATAHEADER) + (SizeOf(TRect) * MaxRects), RgnData^);
    if Result > 0 then
    begin
      CombineRgn(Result, Result, h, RGN_OR);
      DeleteObject(h);
    end
    else
      Result :=h;
  finally
    FreeMem(RgnData, SizeOf(RGNDATAHEADER) + (SizeOf(TRect) * MaxRects));
  end;
end;

//-------------------------------------
procedure Register;
begin
  RegisterComponents('sfCmpnnt', [TTransBtn]);
end;

end.
